﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Dokan;
using System.IO;
using System.Collections;
using MadWizard.WinUSBNet;
using System.Threading;

namespace freemyipod
{
    class emCOREfs : DokanOperations
    {
        USBDevice dev;
        USBPipe coutep;
        USBPipe cinep;
        USBPipe doutep;
        USBPipe dinep;
        uint maxcin;
        uint maxcout;
        uint maxdin;
        uint maxdout;
        bool connected;
        Hashtable openfiles;
        Hashtable opendirs;
        Mutex mutex;

        public emCOREfs()
        {
            connected = false;
            mutex = new Mutex();
            openfiles = new Hashtable();
            opendirs = new Hashtable();
            try
            {
                connect();
            }
            catch (Exception e)
            {
                disconnect();
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
            }
        }

        private void connect()
        {
            if (connected) return;
            Console.WriteLine("Connecting...");
            dev = USBDevice.GetSingleDevice("2084a03c-04b1-4acc-9236-69fe2e7d5770");
            coutep = dev.Interfaces.First().Pipes.ElementAt(0);
            cinep = dev.Interfaces.First().Pipes.ElementAt(1);
            doutep = dev.Interfaces.First().Pipes.ElementAt(2);
            dinep = dev.Interfaces.First().Pipes.ElementAt(3);
            connected = true;
            byte[] data = cmd(new uint[] { 1, 1, 0, 0 }, 16);
            maxcout = BitConverter.ToUInt16(data, 4);
            maxcin = BitConverter.ToUInt16(data, 6);
            maxdout = BitConverter.ToUInt32(data, 8);
            maxdin = BitConverter.ToUInt32(data, 12);
            openfiles = new Hashtable();
            opendirs = new Hashtable();
            closeall();
            fat_enable_flushing(false);
        }

        private void disconnect()
        {
            Console.WriteLine("Disconnected");
            coutep = null;
            cinep = null;
            doutep = null;
            dinep = null;
            dev = null;
            connected = false;
        }

        private int rc(int value)
        {
/*            if (value == 0) Console.WriteLine("Returning OK");
            else if (value == -1) Console.WriteLine("Returning GENERIC_ERROR");
            else if (value == -DokanNet.ERROR_ACCESS_DENIED) Console.WriteLine("Returning ERROR_ACCESS_DENIED");
            else if (value == -DokanNet.ERROR_ALREADY_EXISTS) Console.WriteLine("Returning ERROR_ALREADY_EXISTS");
            else if (value == -DokanNet.ERROR_FILE_EXISTS) Console.WriteLine("Returning ERROR_FILE_EXISTS");
            else if (value == -DokanNet.ERROR_FILE_NOT_FOUND) Console.WriteLine("Returning ERROR_FILE_NOT_FOUND");
            else if (value == -DokanNet.ERROR_INVALID_NAME) Console.WriteLine("Returning ERROR_INVALID_NAME");
            else if (value == -DokanNet.ERROR_PATH_NOT_FOUND) Console.WriteLine("Returning ERROR_PATH_NOT_FOUND");
            else if (value == -DokanNet.ERROR_SHARING_VIOLATION) Console.WriteLine("Returning ERROR_SHARING_VIOLATION");
            else Console.WriteLine("Returning " + value);
*/            return value;
        }

        private byte[] cmd(uint[] args, byte[] payload, uint payloadoffs, uint payloadlength, uint responselength)
        {
            mutex.WaitOne();
            byte[] data;
            try
            {
                connect();
                data = new byte[args.Length * 4 + payloadlength];
                for (int i = 0; i < args.Length; i++)
                    Array.Copy(BitConverter.GetBytes(args[i]), 0, data, i * 4, 4);
                if (payload != null) Array.Copy(payload, payloadoffs, data, args.Length * 4, payloadlength);
/*                Console.Write("OUT");
                for (int i = 0; i < data.Length; i++) Console.Write(" {0:X2}", data[i]);
                Console.WriteLine();
*/                coutep.Write(data);
                data = new byte[responselength];
                int reallength = cinep.Read(data);
/*                Console.Write("IN");
                for (int i = 0; i < reallength; i++) Console.Write(" {0:X2}", data[i]);
                Console.WriteLine();
*/                if (reallength != responselength)
                    throw new Exception("Invalid response length (expected " + responselength + ", got " + reallength + ")");
            }
            catch (Exception e)
            {
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                disconnect();
                mutex.ReleaseMutex();
                throw;
            }
            mutex.ReleaseMutex();
            uint result = BitConverter.ToUInt32(data, 0);
            if (result != 1) throw new Exception("Bad protocol level result: " + result);
            return data;
        }

        private byte[] cmd(uint[] args, byte[] payload, uint responselength)
        {
            return cmd(args, payload, 0, (uint)payload.Length, responselength);
        }

        private byte[] cmd(uint[] args, uint responselength)
        {
            return cmd(args, null, 0, 0, responselength);
        }

        private uint malloc(uint size)
        {
            byte[] data = cmd(new uint[] { 53, 0x10, size, 0 }, 16);
            uint addr = BitConverter.ToUInt32(data, 4);
            if (addr == 0) throw new Exception("Out of device memory");
            return addr;
        }

        private void free(uint addr)
        {
            cmd(new uint[] { 56, addr, 0, 0 }, 16);
        }

        struct split
        {
            public uint head;
            public uint body;
            public uint tail;
        };

        private split alignsplit(uint addr, uint size, uint blksize)
        {
            split result = new split();
            if (size <= blksize)
            {
                result.head = size;
                result.body = 0;
                result.tail = 0;
                return result;
            }
            uint end = addr + size;
            uint bodyaddr;
            if ((addr & 0xf) != 0) bodyaddr = (addr + Math.Min(size, blksize)) & 0xfffffff0;
            else bodyaddr = addr;
            result.head = bodyaddr - addr;
            uint tailaddr;
            if (((size - result.head) & 0xf) != 0) tailaddr = (end - Math.Min(end - bodyaddr, blksize) + 0xf) & 0xfffffff0;
            else tailaddr = end;
            result.tail = end - tailaddr;
            result.body = tailaddr - bodyaddr;
            return result;
        }

        private void upload(uint addr, byte[] data, uint offset, uint length)
        {
            split size = alignsplit(addr, length, maxcout - 16);
            mutex.WaitOne();
            try
            {
                if (size.head != 0) cmd(new uint[] { 5, addr, size.head, 0 }, data, offset, size.head, 16);
                uint offs = size.head;
                while (size.body > 0)
                {
                    try
                    {
                        uint bytes = Math.Min(size.body, maxdout);
                        cmd(new uint[] { 7, addr + offs, bytes, 0 }, 16);
                        doutep.Write(data, (int)(offs + offset), (int)bytes);
                        offs += bytes;
                        size.body -= bytes;
                    }
                    catch
                    {
                        disconnect();
                        throw;
                    }
                }
                if (size.tail != 0) cmd(new uint[] { 5, addr + offs, size.tail, 0 }, data, offs + offset, size.tail, 16);
            }
            finally
            {
                mutex.ReleaseMutex();
            }
        }

        private void upload(uint addr, byte[] data)
        {
            upload(addr, data, 0, (uint)data.Length);
        }

        private void upload(byte[] data, uint offset, uint length)
        {
            upload(malloc(length), data, offset, length);
        }

        private void upload(byte[] data)
        {
            upload(data, 0, (uint)data.Length);
        }

        private byte[] download(uint addr, uint length)
        {
            split size = alignsplit(addr, length, maxcin - 16);
            byte[] result = new byte[length];
            mutex.WaitOne();
            try
            {
                if (size.head != 0)
                    Array.Copy(cmd(new uint[] { 4, addr, size.head, 0 }, 16 + size.head), 16, result, 0, (int)size.head);
                uint offs = size.head;
                while (size.body > 0)
                {
                    try
                    {
                        uint bytes = Math.Min(size.body, maxdin);
                        cmd(new uint[] { 6, addr, bytes, 0 }, 16);
                        dinep.Read(result, (int)offs, (int)bytes);
                        offs += bytes;
                        size.body -= bytes;
                    }
                    catch
                    {
                        disconnect();
                        throw;
                    }
                }
                if (size.tail != 0)
                    Array.Copy(cmd(new uint[] { 4, addr + offs, size.tail, 0 }, 16 + size.tail), 16, result, offs, (int)size.tail);
            }
            finally
            {
                mutex.ReleaseMutex();
            }
            return result;
        }

        struct fatsize
        {
            public uint kb_total;
            public uint kb_free;
        };

        private fatsize getfatsize()
        {
            fatsize result = new fatsize();
            byte[] data = cmd(new uint[] { 59, 0, 0, 0 }, 16);
            result.kb_total = BitConverter.ToUInt32(data, 4);
            result.kb_free = BitConverter.ToUInt32(data, 8);
            return result;
        }

        private uint opendir(string dirname)
        {
            byte[] data = cmd(new uint[] { 42, 0, 0, 0 }, ASCIIEncoding.ASCII.GetBytes(dirname + "\0"), 16);
            int result = BitConverter.ToInt32(data, 4);
            if (result <= 0) throw new Exception("Failed to open directory \"" + dirname + "\": " + result);
            return (uint)result;
        }

        private void closedir(uint handle)
        {
            byte[] data = cmd(new uint[] { 44, handle, 0, 0 }, 16);
            int result = BitConverter.ToInt32(data, 4);
            if (result < 0) throw new Exception("Failed to close directory handle " + handle + ": " + result);
        }

        private uint openfile(string filename, uint mode)
        {
            byte[] data = cmd(new uint[] { 30, mode, 0, 0 }, ASCIIEncoding.ASCII.GetBytes(filename + "\0"), 16);
            int result = BitConverter.ToInt32(data, 4);
            if (result == -1) throw new InvalidDataException();
            if (result == -4) throw new DirectoryNotFoundException();
            if (result == -5) throw new InvalidDataException();
            if (result == -7) throw new FileNotFoundException();
            if (result == -8) throw new ArgumentException();
            if (result <= 0) throw new Exception("Failed to open file \"" + filename + "\": " + result);
            return (uint)result;
        }

        private void closefile(uint handle)
        {
            byte[] data = cmd(new uint[] { 37, handle, 0, 0 }, 16);
            int result = BitConverter.ToInt32(data, 4);
            if (result < 0) throw new Exception("Failed to close file handle " + handle + ": " + result);
        }

        private void sync(uint handle)
        {
            byte[] data = cmd(new uint[] { 36, handle, 0, 0 }, 16);
            int result = BitConverter.ToInt32(data, 4);
            if (result < 0) throw new Exception("Failed to sync file handle " + handle + ": " + result);
        }

        private uint seek(uint handle, uint offset, uint whence)
        {
            byte[] data = cmd(new uint[] { 34, handle, offset, whence }, 16);
            int result = BitConverter.ToInt32(data, 4);
            if (result < 0) throw new Exception("Failed to seek file handle " + handle + " to " + offset + " (mode " + whence + "): " + result);
            return (uint)result;
        }

        private uint truncate(uint handle, uint size)
        {
            byte[] data = cmd(new uint[] { 35, handle, size, 0 }, 16);
            int result = BitConverter.ToInt32(data, 4);
            if (result == -11) throw new ArgumentException();
            if (result < 0) throw new Exception("Failed to truncate file handle " + handle + " to " + size + " bytes: " + result);
            return (uint)result;
        }

        private uint fread(uint handle, uint addr, uint size)
        {
            byte[] data = cmd(new uint[] { 32, handle, addr, size }, 16);
            int result = BitConverter.ToInt32(data, 4);
            if (result < 0) throw new Exception("Failed to read " + size + " bytes from file handle " + handle + " to " + addr + ": " + result);
            return (uint)result;
        }

        private uint fwrite(uint handle, uint addr, uint size)
        {
            byte[] data = cmd(new uint[] { 33, handle, addr, size }, 16);
            int result = BitConverter.ToInt32(data, 4);
            if (result < 0) throw new Exception("Failed to write " + size + " bytes from file handle " + handle + " to " + addr + ": " + result);
            return (uint)result;
        }

        private void close(string filename)
        {
            if (openfiles.Contains(filename))
            {
                closefile((uint)openfiles[filename]);
                openfiles.Remove(filename);
            }
            if (opendirs.Contains(filename))
            {
                closedir((uint)opendirs[filename]);
                opendirs.Remove(filename);
            }
        }

        private void unlink(string filename)
        {
            byte[] data = cmd(new uint[] { 40, 0, 0, 0 }, ASCIIEncoding.ASCII.GetBytes(filename + "\0"), 16);
            int result = BitConverter.ToInt32(data, 4);
            if (result == -11) throw new InvalidDataException();
            if (result == -41) throw new DirectoryNotFoundException();
            if (result == -51) throw new InvalidDataException();
            if (result == -71) throw new FileNotFoundException();
            if (result == -81) throw new ArgumentException();
            if (result < 0) throw new Exception("Failed to unlink \"" + filename + "\": " + result);
        }

        private void rename(string filename, string newname)
        {
            byte[] names = ASCIIEncoding.ASCII.GetBytes(filename.PadRight(248, '\0') + newname + "\0");
            byte[] data = cmd(new uint[] { 41, 0, 0, 0 }, names, 16);
            int result = BitConverter.ToInt32(data, 4);
            if (result == -1) throw new ArgumentException();
            if (result == -12) throw new InvalidDataException();
            if (result == -42) throw new DirectoryNotFoundException();
            if (result == -52) throw new InvalidDataException();
            if (result == -72) throw new FileNotFoundException();
            if (result == -82) throw new ArgumentException();
            if (result == -3) throw new InvalidDataException();
            if (result == -4) throw new InvalidDataException();
            if (result == -5) throw new DirectoryNotFoundException();
            if (result == -16) throw new ArgumentException();
            if (result == -17) throw new ArgumentException();
            if (result < 0) throw new Exception("Failed to rename \"" + filename + "\" to \"" + newname + "\": " + result);
        }

        private void mkdir(string dirname)
        {
            byte[] data = cmd(new uint[] { 47, 0, 0, 0 }, ASCIIEncoding.ASCII.GetBytes(dirname + "\0"), 16);
            int result = BitConverter.ToInt32(data, 4);
            if (result == -2) throw new DirectoryNotFoundException();
            if (result == -3) throw new InvalidDataException();
            if (result == -4) throw new ArgumentException();
            if (result < 0) throw new Exception("Failed to create directory \"" + dirname + "\": " + result);
        }

        private void rmdir(string dirname)
        {
            byte[] data = cmd(new uint[] { 48, 0, 0, 0 }, ASCIIEncoding.ASCII.GetBytes(dirname + "\0"), 16);
            int result = BitConverter.ToInt32(data, 4);
            if (result == -1) throw new DirectoryNotFoundException();
            if (result == -2) throw new ArgumentException();
            if (result < 0) throw new Exception("Failed to remove directory \"" + dirname + "\": " + result);
        }

        struct direntry {
            public string name;
            public uint attributes;
            public uint size;
            public uint startcluster;
            public uint wrtdate;
            public uint wrttime;
        };

        private direntry readdir(uint handle)
        {
            direntry entry = new direntry();
            byte[] data = cmd(new uint[] { 43, handle, 0, 0 }, 16);
            int result = BitConverter.ToInt32(data, 12);
            if (result <= 0) throw new Exception("Failed to read directory entry of handle " + handle + ": " + result);
            uint version = BitConverter.ToUInt32(data, 4);
            if (version != 1) throw new Exception("Unknown dirent struct version: " + version);
            uint maxpath = BitConverter.ToUInt32(data, 8);
            data = download((uint)result, maxpath + 16);
            entry.name = ASCIIEncoding.ASCII.GetString(data, 0, (int)maxpath);
            try
            {
                entry.name = entry.name.Substring(0, entry.name.IndexOf("\0"));
            }
            catch
            {
            }
            entry.attributes = BitConverter.ToUInt32(data, (int)maxpath);
            entry.size = BitConverter.ToUInt32(data, (int)maxpath + 4);
            entry.startcluster = BitConverter.ToUInt32(data, (int)maxpath + 8);
            entry.wrtdate = BitConverter.ToUInt16(data, (int)maxpath + 12);
            entry.wrttime = BitConverter.ToUInt16(data, (int)maxpath + 14);
            return entry;
        }

        private void closeall()
        {
            mutex.WaitOne();
            try
            {
                openfiles = new Hashtable();
                opendirs = new Hashtable();
                cmd(new uint[] { 38, 0, 0, 0 }, 16);
                cmd(new uint[] { 45, 0, 0, 0 }, 16);
                cmd(new uint[] { 57, 0, 0, 0 }, 16);
            }
            finally
            {
                mutex.ReleaseMutex();
            }
        }

        private void fat_enable_flushing(bool state)
        {
            cmd(new uint[] { 58, state ? (uint)1 : (uint)0, 0, 0 }, 16);
        }

        public int CreateFile(String filename, FileAccess access, FileShare share,
            FileMode mode, FileOptions options, DokanFileInfo info)
        {
//            Console.WriteLine("CreateFile(\"" + filename + "\", " + access + ", " + share + ", " + mode + ", " + options + ")");
            filename = filename.Replace("\\", "/");
            mutex.WaitOne();
            if (openfiles.Contains(filename))
            {
                mutex.ReleaseMutex();
                return rc(-DokanNet.ERROR_SHARING_VIOLATION);
            }
            try
            {
                uint flags = 0;
                if (access == FileAccess.Write) flags = 1;
                else if (access == FileAccess.ReadWrite) flags = 2;
                if (mode == FileMode.Append) flags |= 8;
                else if (mode == FileMode.Create) flags |= 0x14;
                else if (mode == FileMode.CreateNew) flags |= 4;
                else if (mode == FileMode.OpenOrCreate) flags |= 4;
                else if (mode == FileMode.Truncate) flags |= 0x10;
                openfiles.Add(filename, openfile(filename, flags));
                mutex.ReleaseMutex();
                return rc(0);
            }
            catch (Exception e)
            {
                mutex.ReleaseMutex();
                if (e is FileNotFoundException) return rc(-DokanNet.ERROR_FILE_NOT_FOUND);
                if (e is DirectoryNotFoundException) return rc(-DokanNet.ERROR_PATH_NOT_FOUND);
                if (e is InvalidDataException) return rc(-DokanNet.ERROR_INVALID_NAME);
                if (e is ArgumentException) return rc(-DokanNet.ERROR_INVALID_NAME);
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                return rc(-DokanNet.ERROR_FILE_NOT_FOUND);
            }
        }

        public int OpenDirectory(String filename, DokanFileInfo info)
        {
//            Console.WriteLine("OpenDirectory(\"" + filename + "\")");
            filename = filename.Replace("\\", "/");
            mutex.WaitOne();
            if (opendirs.Contains(filename))
            {
                mutex.ReleaseMutex();
                return rc(0);
            }
            try
            {
                opendirs.Add(filename, opendir(filename));
                mutex.ReleaseMutex();
                return rc(0);
            }
            catch (Exception e)
            {
                mutex.ReleaseMutex();
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                return rc(-DokanNet.ERROR_FILE_NOT_FOUND);
            }
        }

        public int CreateDirectory(String filename, DokanFileInfo info)
        {
//            Console.WriteLine("CreateDirectory(\"" + filename + "\")");
            filename = filename.Replace("\\", "/");
            try
            {
                mkdir(filename);
                return rc(0);
            }
            catch (Exception e)
            {
                if (e is DirectoryNotFoundException) return rc(-DokanNet.ERROR_PATH_NOT_FOUND);
                if (e is InvalidDataException) return rc(-DokanNet.ERROR_INVALID_NAME);
                if (e is ArgumentException) return rc(-DokanNet.ERROR_ALREADY_EXISTS);
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                return rc(-1);
            }
        }

        public int Cleanup(String filename, DokanFileInfo info)
        {
//            Console.WriteLine("Cleanup(\"" + filename + "\")");
            filename = filename.Replace("\\", "/");
            close(filename);
            return rc(0);
        }

        public int CloseFile(String filename, DokanFileInfo info)
        {
//            Console.WriteLine("CloseFile(\"" + filename + "\")");
            filename = filename.Replace("\\", "/");
            close(filename);
            return rc(0);
        }

        public int ReadFile(String filename, Byte[] buffer, ref uint readBytes,
            long offset, DokanFileInfo info)
        {
//            Console.WriteLine("ReadFile(\"" + filename + "\", " + buffer + ", " + offset + ")");
            filename = filename.Replace("\\", "/");
            mutex.WaitOne();
            try
            {
                if (!openfiles.Contains(filename)) return rc(-1);
                uint file = (uint)openfiles[filename];
                seek(file, (uint)offset, 0);
                uint chunksize = Math.Min((uint)buffer.Length, 65536);
                uint addr = malloc(chunksize);
                readBytes = 0;
                while (readBytes < buffer.Length)
                {
                    uint bytesread = fread(file, addr, chunksize);
                    if (bytesread == 0) break;
                    Array.Copy(download(addr, bytesread), 0, buffer, readBytes, bytesread);
                    readBytes += bytesread;
                }
                free(addr);
                return rc(0);
            }
            catch (Exception e)
            {
                mutex.ReleaseMutex();
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                return rc(-1);
            }
        }

        public int WriteFile(String filename, Byte[] buffer,
            ref uint writtenBytes, long offset, DokanFileInfo info)
        {
//            Console.WriteLine("WriteFile(\"" + filename + "\", " + buffer + ", " + offset + ")");
            filename = filename.Replace("\\", "/");
            mutex.WaitOne();
            try
            {
                if (!openfiles.Contains(filename)) return rc(-1);
                uint file = (uint)openfiles[filename];
                seek(file, (uint)offset, 0);
                uint chunksize = Math.Min((uint)buffer.Length, 65536);
                uint addr = malloc(chunksize);
                writtenBytes = 0;
                while (writtenBytes < buffer.Length)
                {
                    uint maxsize = Math.Min(chunksize, (uint)buffer.Length - writtenBytes);
                    upload(addr, buffer, writtenBytes, maxsize);
                    uint byteswritten = fwrite(file, addr, maxsize);
                    if (byteswritten == 0) break;
                    writtenBytes += byteswritten;
                }
                free(addr);
                return rc(0);
            }
            catch (Exception e)
            {
                mutex.ReleaseMutex();
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                return rc(-1);
            }
        }

        public int FlushFileBuffers(String filename, DokanFileInfo info)
        {
//            Console.WriteLine("FlushFileBuffers(\"" + filename + "\")");
            filename = filename.Replace("\\", "/");
            mutex.WaitOne();
            try
            {
                if (!openfiles.Contains(filename)) return rc(-1);
                uint file = (uint)openfiles[filename];
                fat_enable_flushing(true);
                sync(file);
                fat_enable_flushing(false);
                return rc(0);
            }
            catch (Exception e)
            {
                mutex.ReleaseMutex();
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                return rc(-1);
            }
        }

        public int GetFileInformation(String filename, FileInformation fileinfo, DokanFileInfo info)
        {
//            Console.WriteLine("GetFileInformation(\"" + filename + "\")");
            filename = filename.Replace("\\", "/");
            if (filename == "/")
            {
                fileinfo.Attributes = FileAttributes.Directory;
                fileinfo.CreationTime = new DateTime(1970, 1, 1);
                fileinfo.LastAccessTime = new DateTime(1970, 1, 1);
                fileinfo.LastWriteTime = new DateTime(1970, 1, 1);
                fileinfo.Length = 0;
                fileinfo.FileName = null;
                return rc(0);
            }
            if (filename.Substring(filename.Length - 1) == "/")
                filename = filename.Substring(0, filename.Length - 1);
            int lastslash = filename.LastIndexOf("/");
            string dirname = filename.Substring(0, lastslash + 1);
            string basename = filename.Substring(lastslash + 1);
            try
            {
                uint handle;
                bool close;
                mutex.WaitOne();
                if (opendirs.Contains(dirname))
                {
                    handle = (uint)opendirs[dirname];
                    close = false;
                }
                else
                {
                    handle = opendir(dirname);
                    close = true;
                }
                while (true)
                {
                    direntry entry;
                    try
                    {
                        entry = readdir(handle);
                    }
                    catch
                    {
                        break;
                    }
                    if (entry.name.Equals(basename, StringComparison.OrdinalIgnoreCase))
                    {
                        fileinfo.Attributes = (entry.attributes & 0x10) != 0 ? FileAttributes.Directory : FileAttributes.Normal;
                        fileinfo.CreationTime = new DateTime(1970, 1, 1);
                        fileinfo.LastAccessTime = new DateTime(1970, 1, 1);
                        fileinfo.LastWriteTime = new DateTime(1970, 1, 1);
                        fileinfo.Length = (entry.attributes & 0x10) != 0 ? 0 : entry.size;
                        fileinfo.FileName = entry.name;
                        if (close) closedir(handle);
                        mutex.ReleaseMutex();
                        return rc(0);
                    }
                }
                if (close) closedir(handle);
                mutex.ReleaseMutex();
                return rc(-DokanNet.ERROR_FILE_NOT_FOUND);
            }
            catch (Exception e)
            {
                mutex.ReleaseMutex();
                if (e is DirectoryNotFoundException) return rc(-DokanNet.ERROR_PATH_NOT_FOUND);
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                return rc(-1);
            }
        }

        public int FindFiles(String filename, ArrayList files, DokanFileInfo info)
        {
//            Console.WriteLine("FindFiles(\"" + filename + "\")");
            filename = filename.Replace("\\", "/");
            try
            {
                uint handle;
                bool close;
                mutex.WaitOne();
                if (opendirs.Contains(filename))
                {
                    handle = (uint)opendirs[filename];
                    close = false;
                }
                else
                {
                    handle = opendir(filename);
                    close = true;
                }
                while (true)
                {
                    direntry entry;
                    try
                    {
                        entry = readdir(handle);
                    }
                    catch
                    {
                        break;
                    }
                    FileInformation fi = new FileInformation();
                    fi.Attributes = (entry.attributes & 0x10) != 0 ? FileAttributes.Directory : FileAttributes.Normal;
                    fi.CreationTime = new DateTime(1970, 1, 1);
                    fi.LastAccessTime = new DateTime(1970, 1, 1);
                    fi.LastWriteTime = new DateTime(1970, 1, 1);
                    fi.Length = (entry.attributes & 0x10) != 0 ? 0 : entry.size;
                    fi.FileName = entry.name;
                    files.Add(fi);
                }
                if (close) closedir(handle);
                mutex.ReleaseMutex();
                return rc(0);
            }
            catch (Exception e)
            {
                mutex.ReleaseMutex();
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                return rc(-1);
            }
        }

        public int SetFileAttributes(String filename, FileAttributes attr, DokanFileInfo info)
        {
//            Console.WriteLine("SetFileAttributes(\"" + filename + "\", " + attr + ")");
            filename = filename.Replace("\\", "/");
            return rc(-1);
        }

        public int SetFileTime(String filename, DateTime ctime,
                DateTime atime, DateTime mtime, DokanFileInfo info)
        {
//            Console.WriteLine("SetFileTime(\"" + filename + "\", " + ctime + ", " + atime + ", " + mtime + ")");
            filename = filename.Replace("\\", "/");
            return rc(-1);
        }

        public int DeleteFile(String filename, DokanFileInfo info)
        {
//            Console.WriteLine("DeleteFile(\"" + filename + "\")");
            filename = filename.Replace("\\", "/");
            try
            {
                unlink(filename);
                return rc(0);
            }
            catch (Exception e)
            {
                if (e is ArgumentException) return rc(-DokanNet.ERROR_ALREADY_EXISTS);
                if (e is InvalidDataException) return rc(-DokanNet.ERROR_INVALID_NAME);
                if (e is FileNotFoundException) return rc(-DokanNet.ERROR_FILE_NOT_FOUND);
                if (e is DirectoryNotFoundException) return rc(-DokanNet.ERROR_PATH_NOT_FOUND);
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                return rc(-1);
            }
        }

        public int DeleteDirectory(String filename, DokanFileInfo info)
        {
//            Console.WriteLine("DeleteDirectory(\"" + filename + "\")");
            filename = filename.Replace("\\", "/");
            try
            {
                rmdir(filename);
                return rc(0);
            }
            catch (Exception e)
            {
                if (e is DirectoryNotFoundException) return rc(-DokanNet.ERROR_PATH_NOT_FOUND);
                if (e is ArgumentException) return rc(-DokanNet.ERROR_ALREADY_EXISTS);
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                return rc(-1);
            }
        }

        public int MoveFile(String filename, String newname, bool replace, DokanFileInfo info)
        {
//            Console.WriteLine("MoveFile(\"" + filename + "\", \"" + newname + "\", " + replace + ")");
            filename = filename.Replace("\\", "/");
            newname = newname.Replace("\\", "/");
            try
            {
                if (replace)
                    try
                    {
                        unlink(newname);
                    }
                    catch
                    {
                    }
                rename(filename, newname);
                return rc(0);
            }
            catch (Exception e)
            {
                if (e is ArgumentException) return rc(-DokanNet.ERROR_ALREADY_EXISTS);
                if (e is InvalidDataException) return rc(-DokanNet.ERROR_INVALID_NAME);
                if (e is FileNotFoundException) return rc(-DokanNet.ERROR_FILE_NOT_FOUND);
                if (e is DirectoryNotFoundException) return rc(-DokanNet.ERROR_PATH_NOT_FOUND);
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                return rc(-1);
            }
        }

        public int SetEndOfFile(String filename, long length, DokanFileInfo info)
        {
//            Console.WriteLine("SetEndOfFile(\"" + filename + "\", " + length + ")");
            filename = filename.Replace("\\", "/");
            mutex.WaitOne();
            try
            {
                if (!openfiles.Contains(filename)) return rc(-1);
                uint file = (uint)openfiles[filename];
                truncate(file, (uint)length);
                return rc(0);
            }
            catch (Exception e)
            {
                mutex.ReleaseMutex();
                if (e is ArgumentException) return rc(-1);
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                return rc(-1);
            }
        }

        public int SetAllocationSize(String filename, long length, DokanFileInfo info)
        {
//            Console.WriteLine("SetAllocationSize(\"" + filename + "\", " + length + ")");
            filename = filename.Replace("\\", "/");
            return rc(-1);
        }

        public int LockFile(String filename, long offset, long length, DokanFileInfo info)
        {
//            Console.WriteLine("LockFile(\"" + filename + "\", " + offset + ", " + length + ")");
            filename = filename.Replace("\\", "/");
            return rc(0);
        }

        public int UnlockFile(String filename, long offset, long length, DokanFileInfo info)
        {
//            Console.WriteLine("UnlockFile(\"" + filename + "\", " + offset + ", " + length + ")");
            filename = filename.Replace("\\", "/");
            return rc(0);
        }

        public int GetDiskFreeSpace(ref ulong freeBytesAvailable, ref ulong totalBytes,
            ref ulong totalFreeBytes, DokanFileInfo info)
        {
//            Console.WriteLine("GetDiskFreeSpace()");
            try
            {
                fatsize size = getfatsize();
                freeBytesAvailable = ((ulong)size.kb_free) * 1024;
                totalBytes = ((ulong)size.kb_total) * 1024;
                totalFreeBytes = ((ulong)size.kb_free) * 1024;
                return rc(0);
            }
            catch (Exception e)
            {
                Console.Error.Write(e.Message + "\n" + e.StackTrace + "\n\n");
                return rc(-1);
            }
        }

        public int Unmount(DokanFileInfo info)
        {
//            Console.WriteLine("Unmount()");
            closeall();
            fat_enable_flushing(true);
            return rc(0);
        }

        static void Main(string[] args)
        {
            DokanOptions opt = new DokanOptions();
            opt.ThreadCount = 1;
            opt.MountPoint = "X:\\";
            opt.VolumeLabel = "emCOREfs";
            int status = DokanNet.DokanMain(opt, new emCOREfs());
            switch (status)
            {
                case DokanNet.DOKAN_DRIVE_LETTER_ERROR:
                    Console.WriteLine("Drvie letter error");
                    break;
                case DokanNet.DOKAN_DRIVER_INSTALL_ERROR:
                    Console.WriteLine("Driver install error");
                    break;
                case DokanNet.DOKAN_MOUNT_ERROR:
                    Console.WriteLine("Mount error");
                    break;
                case DokanNet.DOKAN_START_ERROR:
                    Console.WriteLine("Start error");
                    break;
                case DokanNet.DOKAN_ERROR:
                    Console.WriteLine("Unknown error");
                    break;
                case DokanNet.DOKAN_SUCCESS:
                    Console.WriteLine("Success");
                    break;
                default:
                    Console.WriteLine("Unknown status: %d", status);
                    break;

            }
        }
    }
}
